@@ -78,6 +78,49 @@ magnifier_x_error_handler (Display	 *display,
 	return 0;
 }
 
+GdkFilterReturn
+magnifier_expose_filter (GdkXEvent *xevent, GdkEvent *event, gpointer data)
+{
+	Magnifier *magnifier = data;
+	GList *list;
+
+	if (event->any.type == GDK_EXPOSE) 
+	{
+		GNOME_Magnifier_RectBounds rect_bounds;
+		GdkRectangle rect;
+		gdk_region_get_clipbox (event->expose.region, &rect);
+		rect_bounds.x1 = rect.x;
+		rect_bounds.y1 = rect.y;
+		rect_bounds.x2 = rect.x + rect.width;
+		rect_bounds.y2 = rect.y + rect.height;
+		
+		g_assert (magnifier);
+		list = magnifier->zoom_regions;
+		
+		while (list) 
+		{
+			/* propagate the expose events to the zoom regions */
+			ZoomRegion *zoom_region;
+			CORBA_Environment ev;
+			zoom_region = list->data;
+			CORBA_exception_init (&ev);
+			GNOME_Magnifier_ZoomRegion_markDirty (BONOBO_OBJREF (zoom_region),
+							      &rect_bounds,
+							      &ev);
+			list = g_list_next (list);
+		}
+		return GDK_FILTER_TRANSLATE;
+	}
+	return GDK_FILTER_CONTINUE;
+}
+
+static void
+magnifier_set_expose_listener (Magnifier *magnifier, GdkWindow *root)
+{
+	/* todo: remove listener from 'old' root if one was already present */
+	gdk_window_add_filter (root, magnifier_expose_filter, NULL);
+}
+
 static void
 magnifier_realize (GtkWidget *widget)
 {
@@ -124,6 +167,7 @@ magnifier_realize (GtkWidget *widget)
 	 */
 
 	XSetErrorHandler (magnifier_x_error_handler);
+
 	fprintf (stderr, "Realize complete.\n");
 }
 
@@ -259,11 +303,12 @@ magnifier_transform_cursor (Magnifier *magnifier)
 			magnifier->priv->w->window,
 			size, size,
 			1);
-		gdk_pixbuf_render_to_drawable (scaled_cursor_pixbuf,
-					       magnifier->priv->cursor,
-					       cgc,
-					       0, 0, 0, 0, size, size,
-					       GDK_RGB_DITHER_NONE, 0, 0 );
+
+		gdk_draw_pixbuf (magnifier->priv->cursor,
+				 cgc,
+				 scaled_cursor_pixbuf,
+				 0, 0, 0, 0, size, size,
+				 GDK_RGB_DITHER_NONE, 0, 0 );
 		scaled_mask_pixbuf = gdk_pixbuf_add_alpha (
 			scaled_mask_pixbuf, True, 0, 0, 0);
 		gdk_pixbuf_render_threshold_alpha (scaled_mask_pixbuf,
@@ -566,6 +611,9 @@ impl_magnifier_set_source_display (PortableServer_Servant servant,
 				ev,
 				magnifier);
 
+	/* attach listener to root window of source display */
+	magnifier_set_expose_listener (magnifier, magnifier_get_root (magnifier));
+
 	bonobo_arg_release (arg);
 }
 
@@ -650,6 +698,13 @@ impl_magnifier_add_zoom_region (PortableServer_Servant servant,
 {
 	Magnifier *magnifier = MAGNIFIER (bonobo_object_from_servant (servant));
 
+	if (magnifier->zoom_regions == NULL) 
+	{
+		/* first zoom region */
+		/* add an expose handler to root window; for XSun/XVirtualScreen 'dirty region' support */
+		magnifier_set_expose_listener (magnifier, magnifier_get_root (magnifier));
+	}
+
 	/* FIXME: this needs proper lifecycle management */
 	magnifier->zoom_regions = g_list_append (magnifier->zoom_regions, region);
 
